import math
from queue import Queue
import random
import time
from collections import defaultdict
import threading
from threading import Lock
from typing import Any, Dict, List, Literal

from cachetools import TTLCache, cached
import numpy as np
import pandas as pd
import polars as pl
import io
import re
import os
import zipfile
import requests
import platform

from quantplay.exception.exceptions import (
    QuantplayOrderPlacementException,
    InvalidArgumentException,
    StaleDataFound,
    FeatureNotSupported,
)
from quantplay.model.broker import (
    ExchangeType,
    ModifyOrderRequest,
    OrderType,
    ProductType,
    TransactionType,
    UserBrokerProfileResponse,
)
from quantplay.utils.constant import Constants
from quantplay.utils.exchange import Market as MarketConstants
from quantplay.utils.number_utils import NumberUtils
from quantplay.utils.constant import timeit
import traceback
import copy
from datetime import datetime, timedelta
from quantplay.utils.pickle_utils import PickleUtils, InstrumentData
from quantplay.wrapper.aws.s3 import S3Utils

logger = Constants.logger


class Broker:
    def __init__(self):
        self.instrument_data = pd.DataFrame()
        self.instrument_id_to_symbol_map = dict()
        self.instrument_id_to_exchange_map = dict()
        self.instrument_id_to_security_type_map = dict()
        self.exchange_symbol_to_instrument_id_map = defaultdict(dict)

        self.quantplay_symbol_map: Dict[str, str] = {}
        self.broker_symbol_map: Dict[str, str] = {}

        self.order_type_sl = "SL"
        self.nfo_exchange = "NFO"

        self.username = None
        self.nickname = None
        self.broker_name = None

        self.order_updates = Queue()
        self.user_id: str | None = None

        self.orders_schema: Dict = {
            "user_id": pl.String,
            "order_id": pl.String,
            "exchange": pl.String,
            "tradingsymbol": pl.String,
            "product": pl.String,
            "variety": pl.String,
            "status": pl.String,
            "order_type": pl.String,
            "transaction_type": pl.String,
            "tag": pl.String,
            "price": pl.Float64,
            "average_price": pl.Float64,
            "trigger_price": pl.Float64,
            "ltp": pl.Float64,
            "order_timestamp": pl.Datetime(time_unit="ms"),
            "token": pl.Int64,
            "quantity": pl.Int64,
            "pending_quantity": pl.Int64,
            "filled_quantity": pl.Int64,
            "status_message": pl.String,
            "status_message_raw": pl.String,
        }

        self.positions_schema: Dict = {
            "tradingsymbol": pl.String,
            "sell_value": pl.Float64,
            "average_price": pl.Float64,
            "quantity": pl.Int64,
            "buy_value": pl.Float64,
            "product": pl.String,
            "ltp": pl.Float64,
            "pnl": pl.Float64,
            "token": pl.Int64,
            "exchange": pl.String,
            "sell_quantity": pl.Int64,
            "option_type": pl.String,
            "buy_quantity": pl.Int64,
        }

        self.holidings_schema: Dict = {
            "tradingsymbol": pl.String,
            "exchange": pl.String,
            "token": pl.Int64,
            "isin": pl.String,
            "quantity": pl.Int64,
            "pledged_quantity": pl.Int64,
            "average_price": pl.Float64,
            "price": pl.Float64,
            "buy_value": pl.Float64,
            "current_value": pl.Float64,
            "pct_change": pl.Float64,
        }

        self.trigger_pending_status = "TRIGGER PENDING"
        self.lock = Lock()
        self.square_off_lock = Lock()

    def validate_exchange(self, exchange):
        if exchange not in ["NSE", "NFO", "BFO", "BSE"]:
            raise InvalidArgumentException(f"exchange {exchange} not supported")

    def validate_existance(self, input, message):
        if not input:
            raise InvalidArgumentException(message)

    def symbol_attribute(self, exchange, symbol, value):
        try:
            return_value = self.symbol_data[f"{exchange}:{symbol}"][value]
            return return_value
        except KeyError:
            logger.error(
                f"[MISSING_SYMBOL_DATA] for [{exchange}:{symbol}] attribute {value}"
            )
            raise StaleDataFound(f"Couldn't find symbol data for [{exchange}:{symbol}]")

    def initialize_symbol_data(self, save_as: str | None = None) -> None:
        instruments = self.instrument_data
        instruments = instruments.to_dict("records")
        self.symbol_data = {}
        for instrument in instruments:
            exchange = instrument["exchange"]
            tradingsymbol = instrument["broker_symbol"]

            instrument_data = copy.deepcopy(instrument)
            self.symbol_data["{}:{}".format(exchange, tradingsymbol)] = instrument_data

        if save_as:
            PickleUtils.save_data(self.symbol_data, save_as)

    @timeit(MetricName="Broker:initialize_broker_symbol_map")
    def initialize_broker_symbol_map(self):
        self.broker_symbol_map = {}
        for a in self.symbol_data:
            self.broker_symbol_map[self.symbol_data[a]["broker_symbol"]] = (
                self.symbol_data[a]["tradingsymbol"]
            )
        platform_version = platform.python_version().split(".")
        if (
            len(platform_version) >= 2
            and platform_version[0] == "3"
            and platform_version[1] == "8"
        ):
            self.quantplay_symbol_map: Dict[str, str] = {}
            for k in self.broker_symbol_map:
                v = self.broker_symbol_map[k]
                self.quantplay_symbol_map[v] = k
        else:
            self.quantplay_symbol_map = {v: k for k, v in self.broker_symbol_map.items()}

    @timeit(MetricName="Broker:load_instrument")
    def load_instrument(self, file_name: str) -> None:
        try:
            instrument_data_instance = InstrumentData.get_instance()

            if instrument_data_instance is not None:
                self.symbol_data = instrument_data_instance.load_data(file_name)

            Constants.logger.info("[LOADING_INSTRUMENTS] loading data from cache")
        except Exception:
            self.instrument_data = S3Utils.read_csv(
                "quantplay-market-data",
                f"symbol_data/{file_name}.csv",
                read_from_local=False,
            )
            self.initialize_symbol_data(save_as=file_name)

        self.initialize_broker_symbol_map()

    def round_to_tick(self, number):
        return round(number * 20) / 20

    def get_user_id(self):
        return self.user_id

    def populate_instruments(self, instruments):
        """Fetches instruments for all exchanges from the broker
        and stores them in the member attributes.
        """
        Constants.logger.info("populating instruments")
        for instrument in instruments:
            exchange, symbol, instrument_id = (
                instrument.exchange,
                instrument.symbol,
                instrument.instrument_id,
            )
            self.instrument_id_to_symbol_map[instrument_id] = symbol
            self.instrument_id_to_exchange_map[instrument_id] = exchange
            self.instrument_id_to_security_type_map[instrument_id] = (
                instrument.security_type()
            )
            self.exchange_symbol_to_instrument_id_map[exchange][symbol] = instrument_id

    def add_quantplay_fut_tradingsymbol(self):
        seg_condition = [
            (
                (self.instrument_data["instrument"].str.contains("FUT"))
                & (self.instrument_data.instrument != "OPTFUT")
            )
        ]

        tradingsymbol = [
            self.instrument_data.tradingsymbol
            + self.instrument_data.expiry_year
            + self.instrument_data.month
            + "FUT"
        ]

        self.instrument_data.loc[:, "tradingsymbol"] = np.select(
            seg_condition, tradingsymbol, default=self.instrument_data.tradingsymbol
        )

    def add_quantplay_opt_tradingsymbol(self):
        seg_condition = self.instrument_data["strike_price"] > 0
        weekly_option_condition = (
            self.instrument_data.expiry.dt.month
            == self.instrument_data.next_expiry.dt.month
        ) & (self.instrument_data.exchange == "NFO")
        month_option_condition = (
            self.instrument_data.expiry.dt.month
            != self.instrument_data.next_expiry.dt.month
        ) | (self.instrument_data.exchange == "MCX")

        self.instrument_data.loc[:, "tradingsymbol"] = np.where(
            seg_condition,
            self.instrument_data.tradingsymbol + self.instrument_data.expiry_year,
            self.instrument_data.tradingsymbol,
        )

        self.instrument_data.loc[:, "tradingsymbol"] = np.where(
            seg_condition & weekly_option_condition,
            self.instrument_data.tradingsymbol + self.instrument_data.week_option_prefix,
            self.instrument_data.tradingsymbol,
        )

        self.instrument_data.loc[:, "tradingsymbol"] = np.where(
            seg_condition & month_option_condition,
            self.instrument_data.tradingsymbol + self.instrument_data.month,
            self.instrument_data.tradingsymbol,
        )

        self.instrument_data.loc[:, "tradingsymbol"] = np.where(
            seg_condition,
            self.instrument_data.tradingsymbol
            + self.instrument_data.strike_price.astype(float)
            .astype(str)
            .str.split(".")
            .str[0],
            self.instrument_data.tradingsymbol,
        )

        self.instrument_data.loc[:, "tradingsymbol"] = np.where(
            seg_condition,
            self.instrument_data.tradingsymbol + self.instrument_data.option_type,
            self.instrument_data.tradingsymbol,
        )

    def get_df_from_zip(self, url):
        response = requests.get(url, timeout=10)
        z = zipfile.ZipFile(io.BytesIO(response.content))

        directory = "/tmp/"
        z.extractall(path=directory)
        file_name = url.split(".txt")[0].split("/")[-1]
        os.system("cp /tmp/{}.txt /tmp/{}.csv".format(file_name, file_name))
        time.sleep(2)
        return pd.read_csv("/tmp/{}.csv".format(file_name))

    def initialize_expiry_fields(self):
        self.instrument_data.loc[:, "tradingsymbol"] = (
            self.instrument_data.instrument_symbol
        )
        self.instrument_data.loc[:, "expiry"] = pd.to_datetime(
            self.instrument_data.instrument_expiry
        )

        self.instrument_data.loc[:, "expiry_year"] = (
            self.instrument_data["expiry"].dt.strftime("%y").astype(str)
        )
        self.instrument_data.loc[:, "month"] = (
            self.instrument_data["expiry"].dt.strftime("%b").str.upper()
        )

        self.instrument_data.loc[:, "month_number"] = (
            self.instrument_data["expiry"].dt.strftime("%m").astype(float).astype(str)
        )
        self.instrument_data.loc[:, "month_number"] = np.where(
            self.instrument_data.month_number == "nan",
            np.nan,
            self.instrument_data.month_number.str.split(".").str[0],
        )

        self.instrument_data.loc[:, "week_option_prefix"] = np.where(
            self.instrument_data.month_number.astype(float) >= 10,
            self.instrument_data.month.str[0]
            + self.instrument_data["expiry"].dt.strftime("%d").astype(str),
            self.instrument_data.month_number
            + self.instrument_data["expiry"].dt.strftime("%d").astype(str),
        )

        self.instrument_data.loc[:, "next_expiry"] = (
            self.instrument_data.expiry + pd.DateOffset(days=7)
        )

    @cached(cache=TTLCache(maxsize=1, ttl=2))
    def cached_orders(self) -> pl.DataFrame:
        return self.orders()

    @cached(cache=TTLCache(maxsize=1, ttl=2))
    def cached_positions(self) -> pl.DataFrame:
        return self.positions()

    def execute_order(
        self,
        tradingsymbol,
        exchange,
        quantity,
        order_type,
        transaction_type,
        stoploss,
        tag,
        product,
        price,
    ):
        upper_circuit = None
        trade_price = copy.deepcopy(price)
        if price is None:
            live_data = self.live_data(exchange=exchange, tradingsymbol=tradingsymbol)
            price = live_data["ltp"]
            upper_circuit = live_data["upper_circuit"]
            trade_price = copy.deepcopy(price)
        try:
            if stoploss is not None:
                if transaction_type == "SELL":
                    sl_transaction_type = "BUY"
                    sl_trigger_price = self.round_to_tick(price * (1 + stoploss))

                    if exchange == "NFO":
                        price = sl_trigger_price * 1.05
                    elif exchange == "NSE":
                        price = sl_trigger_price * 1.01
                    else:
                        raise Exception("{} not supported for trading".format(exchange))

                    sl_price = self.round_to_tick(price)
                elif transaction_type == "BUY":
                    sl_transaction_type = "SELL"
                    sl_trigger_price = self.round_to_tick(price * (1 - stoploss))

                    if exchange == self.nfo_exchange:
                        price = sl_trigger_price * 0.95
                    elif exchange == "NSE":
                        price = sl_trigger_price * 0.99
                    else:
                        raise Exception("{} not supported for trading".format(exchange))

                    sl_price = self.round_to_tick(price)
                else:
                    raise Exception(
                        "Invalid transaction_type {}".format(transaction_type)
                    )

                if upper_circuit is not None and sl_price > upper_circuit:
                    raise Exception(
                        f"[PRICE_BREACHED] {sl_price} is above upper circuit price [{upper_circuit}]"
                    )

                stoploss_order_id = self.place_order(
                    tradingsymbol=tradingsymbol,
                    exchange=exchange,
                    quantity=quantity,
                    order_type=self.order_type_sl,  # type: ignore
                    transaction_type=sl_transaction_type,
                    tag=tag,
                    product=product,
                    price=sl_price,
                    trigger_price=sl_trigger_price,
                )

                if stoploss_order_id is None:
                    Constants.logger.error(
                        "[ORDER_REJECTED] tradingsymbol {}".format(tradingsymbol)
                    )
                    raise QuantplayOrderPlacementException(
                        "Order reject for {}".format(tradingsymbol)
                    )

            if order_type == "MARKET":
                trade_price = 0

            response = self.place_order(
                tradingsymbol=tradingsymbol,
                exchange=exchange,
                quantity=quantity,
                order_type=order_type,
                transaction_type=transaction_type,
                tag=tag,
                product=product,
                price=trade_price,
            )
            return response
        except Exception as e:
            raise e

    """
            Input  : quantplay symbol
            Output : broker symbol
        """

    def get_symbol(self, symbol, exchange=None):
        return symbol

    """
        Input  : quantplay exchange
        Output : broker exchange
    """

    def get_order_type(self, order_type):
        return order_type

    def get_exchange(self, exchange):
        return exchange

    def live_data(self, exchange, tradingsymbol):
        return {
            "ltp": self.get_ltp(exchange, tradingsymbol),
            "upper_circuit": None,
            "lower_circuit": None,
        }

    def basket_margin(self, basket_orders) -> Dict[str, Any]:
        raise FeatureNotSupported("Margin calculator not supported by broker")

    def verify_rms_square_off(self, stoploss, target, keep_hedges=False, ticks=1):
        positions = self.positions()
        pnl = positions["pnl"].sum()

        positions = positions[positions["product"] != "CNC"]
        if keep_hedges:
            positions = positions.filter(
                (~pl.col("option_type").is_in(["CE", "PE"])) & (pl.col("quantity").gt(0))
            )

        if len(positions.filter(pl.col("quantity").ne(0))) == 0:
            return {"should_exit": False, "pnl": pnl}
        # Account level stoploss

        if stoploss is not None and pnl < stoploss:
            if ticks > 0:
                time.sleep(1)
                return self.verify_rms_square_off(stoploss, target, ticks=ticks - 1)
            logger.critical(
                f"[RMS_WARNING] pnl[{pnl}] went below stoploss [{stoploss}] for user {self.profile()}"
            )
            return {"should_exit": True, "pnl": pnl}
        if target is not None and pnl > target:
            if ticks > 0:
                time.sleep(1)
                return self.verify_rms_square_off(stoploss, target, ticks=ticks - 1)
            logger.critical(
                f"[RMS_WARNING] pnl[{pnl}] went above target[{target}] for user {self.profile()}"
            )
            return {"should_exit": True, "pnl": pnl}

        return {"should_exit": False, "pnl": pnl}

    def place_order_quantity(self, quantity, tradingsymbol, exchange):
        lot_size = self.get_lot_size(exchange, tradingsymbol)
        quantity_in_lots = int(quantity / lot_size)

        return quantity_in_lots * lot_size

    def get_product(self, product):
        return product

    def get_lot_size(self, exchange, tradingsymbol):
        tradingsymbol = self.get_symbol(tradingsymbol, exchange=exchange)
        exchange = self.get_exchange(exchange)
        if exchange == "BSE" or exchange == "NSE":
            return 1
        try:
            return int(
                self.symbol_data["{}:{}".format(exchange, tradingsymbol)]["lot_size"]
            )
        except Exception as e:
            logger.error(
                f"[GET_LOT_SIZE] unable to get lot size for {exchange} {tradingsymbol}"
            )
            raise e

    def filter_orders(self, orders, tag=None, status=None):
        if tag:
            orders = orders[orders.tag == tag]

        if status:
            orders = orders[orders.status == status]

        return orders

    def option_symbol(self, underlying_symbol, expiry_date, strike_price, type):
        option_symbol = MarketConstants.INDEX_SYMBOL_TO_DERIVATIVE_SYMBOL_MAP[
            underlying_symbol
        ]
        option_symbol += expiry_date.strftime("%y")

        month_number = str(int(expiry_date.strftime("%m")))
        monthly_option_prefix = expiry_date.strftime("%b").upper()

        if int(month_number) >= 10:
            week_option_prefix = monthly_option_prefix[0]
        else:
            week_option_prefix = month_number
        week_option_prefix += expiry_date.strftime("%d")

        next_expiry = expiry_date + timedelta(days=7)

        if next_expiry.month != expiry_date.month:
            option_symbol += monthly_option_prefix
        else:
            option_symbol += week_option_prefix

        option_symbol += str(int(strike_price))
        option_symbol += type

        return option_symbol

    def exit_all_trigger_orders(
        self,
        tag="ALL",
        symbol_contains=None,
        order_timestamp=None,
        modify_sleep_time=10,
    ):
        stoploss_orders = self.orders()
        stoploss_orders = stoploss_orders.filter(pl.col("status") == "TRIGGER PENDING")

        if len(stoploss_orders) == 0:
            Constants.logger.info("All stoploss orders have been already closed")
            return

        if tag != "ALL":
            stoploss_orders = stoploss_orders.filter(pl.col("tag") == tag)

        if symbol_contains is not None:
            symbol_contains = self.get_symbol(symbol_contains)
            stoploss_orders = stoploss_orders.filter(
                pl.col("tradingsymbol").str.contains(symbol_contains)
            )

        if order_timestamp is not None:
            stoploss_orders = stoploss_orders.with_columns(
                pl.struct(["order_timestamp"])
                .map_elements(
                    lambda x: x["order_timestamp"].replace(second=0),
                    return_dtype=pl.Datetime,
                )
                .alias("order_timestamp")
            )

            stoploss_orders = stoploss_orders.filter(
                pl.col("order_timestamp").dt.strftime("%Y-%m-%d %H:%M:%S")
                == order_timestamp
            )

        if len(stoploss_orders) == 0:
            Constants.logger.info("All stoploss orders have been already closed")
            return

        orders_to_close = list(stoploss_orders["order_id"].unique())

        stoploss_orders = stoploss_orders.to_dicts()
        for stoploss_order in stoploss_orders:
            exchange = stoploss_order["exchange"]
            tradingsymbol = stoploss_order["tradingsymbol"]

            ltp = self.get_ltp(exchange, tradingsymbol)
            stoploss_order["order_type"] = "LIMIT"
            stoploss_order["price"] = self.round_to_tick(ltp)
            stoploss_order["trigger_price"] = None

            self.modify_order(stoploss_order)  # type: ignore
            time.sleep(0.1)

        self.modify_orders_till_complete(orders_to_close, sleep_time=modify_sleep_time)
        Constants.logger.info("All order have been closed successfully")

    def market_protection_price(
        self,
        price: float,
        transaction_type: Literal["BUY", "SELL"],
        market_protection: float = 0.02,
    ) -> float:
        if transaction_type == "BUY":
            price = self.round_to_tick(price * (1 + market_protection))
            return price + 1

        elif transaction_type == "SELL":
            return self.round_to_tick(price * (1 - market_protection))

        return price

    def split_order(self, exchange, tradingsymbol, quantity, max_qty=None):
        max_lots = self.symbol_max_lots(exchange, tradingsymbol)
        lot_size = self.get_lot_size(exchange, tradingsymbol)

        quantity_in_lots = int(quantity / lot_size)
        if max_qty:
            max_lots = max_qty / lot_size

        split_into = int(math.ceil(quantity_in_lots / max_lots))
        split_array = NumberUtils.split(abs(quantity_in_lots), abs(split_into))
        return [a * lot_size for a in split_array]

    def cancel_open_orders(self, tradingsymbols=None):
        open_orders = self.orders()
        open_orders = open_orders.filter(
            pl.col("status").is_in(["OPEN", "TRIGGER PENDING"])
        )

        if tradingsymbols is not None:
            tradingsymbols = list(set(tradingsymbols))
            open_orders = open_orders.filter(
                pl.col("tradingsymbol").is_in(tradingsymbols)
            )

        order_ids = open_orders["order_id"].to_list()
        for order_id in order_ids:
            self.cancel_order(order_id)
            time.sleep(0.1)

    @timeit(MetricName="Broker:square_off_all")
    def square_off_all(
        self,
        dry_run=True,
        contains=None,
        option_type=None,
        sleep_time=0.1,
        keep_hedges=False,
        modify_sleep_time=5,
        max_modification_count=10,
        market_protection=0.02,
    ):
        positions = self.positions()
        positions = positions.filter(pl.col("product") != "CNC")

        if option_type and "option_type" in positions.columns:
            positions = positions.filter(pl.col("option_type") == option_type)

        if contains:
            positions = positions.filter(pl.col("tradingsymbol").str.contains(contains))

        positions = positions.with_columns(
            (pl.col("buy_quantity") - pl.col("sell_quantity")).alias("net_quantity")
        )
        positions = positions.filter(pl.col("net_quantity") != 0)

        if keep_hedges:
            positions = positions.filter(
                ~((pl.col("option_type").is_in(["CE", "PE"])) & (pl.col("quantity") > 0))
            )
        if len(positions) == 0:
            print("Positions are already squared off")
            return []

        positions = positions.with_columns(
            pl.when(pl.col("quantity") < 0)
            .then("BUY")
            .otherwise("SELL")
            .alias("transaction_type")
        )

        positions = positions.with_columns(
            pl.struct(["exchange", "tradingsymbol"])
            .map_elements(
                lambda x: int(self.get_lot_size(x["exchange"], x["tradingsymbol"])),
                return_dtype=pl.Int64,
            )
            .alias("lot_size")
        )

        positions = positions.with_columns(
            pl.struct(["exchange", "tradingsymbol"])
            .map_elements(
                lambda x: int(self.get_ltp(x["exchange"], x["tradingsymbol"])),
                return_dtype=pl.Float64,
            )
            .alias("price")
        )

        positions = positions.to_dicts()
        orders_to_close = []
        for position in positions:
            quantity = abs(position["net_quantity"])
            exchange = position["exchange"]
            tradingsymbol = position["tradingsymbol"]
            transaction_type = position["transaction_type"]

            quantity_in_lots = int(quantity / self.get_lot_size(exchange, tradingsymbol))

            max_lots = self.symbol_max_lots(exchange, tradingsymbol)
            split_into = int(math.ceil(quantity_in_lots / max_lots))
            split_array = NumberUtils.split(abs(quantity_in_lots), abs(split_into))

            for q in split_array:
                orders_to_close.append(
                    {
                        "symbol": tradingsymbol,
                        "exchange": exchange,
                        "transaction_type": transaction_type,
                        "quantity_in_lots": q,
                        "product": position["product"],
                        "price": position["price"],
                    }
                )

        random.shuffle(orders_to_close)
        orders_to_close = sorted(orders_to_close, key=lambda d: d["transaction_type"])

        tradingsymbols = [a["symbol"] for a in orders_to_close]
        if not dry_run:
            self.cancel_open_orders(tradingsymbols)

        orders_placed = []
        for order in orders_to_close:
            quantity = int(
                order["quantity_in_lots"]
                * self.get_lot_size(order["exchange"], order["symbol"])
            )

            order["price"] = self.market_protection_price(
                order["price"],
                order["transaction_type"],
                market_protection=market_protection,
            )
            print(
                order["symbol"],
                order["exchange"],
                order["transaction_type"],
                quantity,
                order["price"],
            )
            if not dry_run:
                order_id = self.place_order(
                    tradingsymbol=order["symbol"],
                    exchange=order["exchange"],
                    quantity=quantity,
                    order_type="LIMIT",
                    transaction_type=order["transaction_type"],
                    tag="killall",
                    product=order["product"],
                    price=order["price"],
                )
                orders_placed.append(str(order_id))
                time.sleep(sleep_time)
        if not dry_run:
            self.modify_orders_till_complete(
                orders_placed,
                sleep_time=modify_sleep_time,
                max_modification_count=max_modification_count,
            )

        return orders_to_close

    def convert_to_event(self, order):
        order = copy.deepcopy(order)
        order["placed_by"] = self.user_id

        order["exchange_order_id"] = order["order_id"]
        order["quantity"] = int(order["quantity"])
        order["tradingsymbol"] = self.get_quantplay_symbol(order["tradingsymbol"])

        order["order_timestamp"] = str(order["order_timestamp"])
        order["update_timestamp"] = str(order["update_timestamp"])

        if "trigger_price" in order and order["trigger_price"] != 0:
            order["trigger_price"] = float(order["trigger_price"])
        else:
            order["trigger_price"] = None
        Constants.logger.info(f"[ORDER_EVENT] {order}")
        if order["status"] == "COMPLETE":
            order["status"] = "OPEN"
            self.order_updates.put(copy.deepcopy(order))
            order["status"] = "COMPLETE"
            time.sleep(1)
        self.order_updates.put(order)

    def get_quantplay_symbol(self, symbol):
        return self.broker_symbol_map[symbol]

    def stream_order_updates(self):
        self.order_log = {}
        while 1:
            try:
                orders = self.orders(add_ltp=False)

                if len(orders) > 0:
                    delta_time = str(
                        datetime.now().replace(microsecond=0) - timedelta(seconds=120)
                    )
                    orders = orders[orders["update_timestamp"] >= delta_time]

                    orders = orders.to_dicts()
                    for order in orders:
                        order_id = order["order_id"]
                        if "status" in order and order["status"].lower() in [
                            "partly executed"
                        ]:
                            continue
                        if order_id not in self.order_log:
                            self.convert_to_event(order)
                        else:
                            update_timestamp = str(order["update_timestamp"])
                            last_log_time = str(
                                self.order_log[order_id]["update_timestamp"]
                            )

                            if update_timestamp != last_log_time:
                                self.convert_to_event(order)
                        self.order_log[order_id] = copy.deepcopy(order)
                time.sleep(3)
            except Exception:
                print(traceback.print_exc())
                print(f"Unable to process order stream for {self.user_id}")
                time.sleep(5)

    def stream_order_data(self):
        th = threading.Thread(target=self.stream_order_updates, daemon=True)
        th.start()

    def underlying_config(self, underlying_symbol, expiry=None):
        if underlying_symbol in ["BANKNIFTY", "BANKEX"]:
            return {"max_lots": 60, "lot_size": 15, "strike_gap": 100}
        elif underlying_symbol == "FINNIFTY":
            return {"max_lots": 45, "lot_size": 40, "strike_gap": 50}
        elif underlying_symbol == "SENSEX":
            return {"max_lots": 100, "lot_size": 10, "strike_gap": 100}
        elif underlying_symbol == "NIFTY":
            return {"max_lots": 72, "lot_size": 25, "strike_gap": 50}
        elif underlying_symbol == "MIDCPNIFTY":
            return {"max_lots": 37, "lot_size": 75, "strike_gap": 25}

        raise Exception(f"Underlying {underlying_symbol} symbol not supported")

    def symbol_max_lots(self, exchange, symbol):
        try:
            if symbol in self.broker_symbol_map:
                symbol = self.broker_symbol_map[symbol]
            if "BANKNIFTY" in symbol and exchange == "NFO":
                return self.underlying_config("BANKNIFTY")["max_lots"]
            elif "FINNIFTY" in symbol and exchange == "NFO":
                return self.underlying_config("FINNIFTY")["max_lots"]
            elif "MIDCPNIFTY" in symbol and exchange == "NFO":
                return self.underlying_config("MIDCPNIFTY")["max_lots"]
            elif "NIFTY" in symbol and exchange == "NFO":
                return self.underlying_config("NIFTY")["max_lots"]
            elif "BANKEX" in symbol and exchange == "BFO":
                return self.underlying_config("BANKEX")["max_lots"]
            elif "SENSEX" in symbol and exchange == "BFO":
                return self.underlying_config("SENSEX")["max_lots"]
            elif exchange == "NSE":
                max_qty = int(500000 / self.get_ltp(exchange, symbol))
                if max_qty == 0:
                    return 1
                return max_qty
            return 36
        except Exception:
            traceback.print_exc()
            logger.error(f"Couldn't compute freeze quantity for {exchange} {symbol}")
            return 25

    def square_off_by_tag(
        self, tag, dry_run=True, sleep_time=0.05
    ) -> List[Dict[str, Any]]:
        self.exit_all_trigger_orders(tag=tag)
        orders = self.orders(tag=tag)

        if len(orders) == 0:
            logger.info(
                f"All positions with tag {tag} are already squared-off for {self.profile()}"
            )
        orders = orders.with_columns(
            pl.when(pl.col("transaction_type") == "BUY")
            .then(-pl.col("filled_quantity"))
            .otherwise(pl.col("filled_quantity"))
            .alias("exit_quantity")
        )

        exit_orders = orders.group_by("tradingsymbol").agg(
            [
                pl.sum("exit_quantity").alias("exit_quantity"),
                pl.col("exchange").first().alias("exchange"),
                pl.col("product").first().alias("product"),
            ]
        )

        orders_to_close: List[Dict[str, Any]] = []
        exit_orders = exit_orders.filter(pl.col("exit_quantity") != 0)
        positions = exit_orders.to_dicts()
        for position in positions:
            exchange = position["exchange"]
            tradingsymbol = position["tradingsymbol"]
            quantity = position["exit_quantity"]

            transaction_type = "SELL"
            if quantity == 0:
                continue
            elif quantity > 0:
                transaction_type = "BUY"

            quantity = abs(quantity)
            quantity_in_lots = int(quantity / self.get_lot_size(exchange, tradingsymbol))

            split_into = int(math.ceil(quantity_in_lots / 25))
            split_array = NumberUtils.split(abs(quantity_in_lots), abs(split_into))

            for q in split_array:
                orders_to_close.append(
                    {
                        "tradingsymbol": tradingsymbol,
                        "exchange": exchange,
                        "transaction_type": transaction_type,
                        "quantity_in_lots": q,
                        "product": position["product"],
                    }
                )

        random.shuffle(orders_to_close)
        orders_to_close = sorted(orders_to_close, key=lambda d: d["transaction_type"])
        for order in orders_to_close:
            tradingsymbol = order["tradingsymbol"]
            exchange = order["exchange"]
            transaction_type = order["transaction_type"]
            product = order["product"]
            quantity = order["quantity_in_lots"] * self.get_lot_size(
                exchange, tradingsymbol
            )
            quantity = self.place_order_quantity(quantity, tradingsymbol, exchange)

            print(tradingsymbol, exchange, transaction_type, quantity)
            if not dry_run:
                self.place_order(
                    tradingsymbol=tradingsymbol,
                    exchange=exchange,
                    quantity=quantity,
                    order_type="MARKET",
                    transaction_type=transaction_type,
                    tag=tag,
                    product=product,
                    price=0,
                )
                time.sleep(sleep_time)

        return orders_to_close

    def add_ltp(self, orders):
        orders.loc[:, "exchange_symbol"] = orders.exchange + ":" + orders.tradingsymbol

        all_symbols = list(orders.exchange_symbol.unique())
        symbol_ltp = {}
        for exchange_symbol in all_symbols:
            ltp = self.get_ltp(
                exchange_symbol.split(":")[0], exchange_symbol.split(":")[1]
            )
            symbol_ltp[exchange_symbol] = ltp
        orders.loc[:, "ltp"] = orders["exchange_symbol"].map(symbol_ltp)

    def modify_price(self, order_id, price, trigger_price=None, order_type=None):
        data = {
            "order_id": order_id,
            "price": price,
            "order_type": order_type,
            "variety": "regular",
            "trigger_price": trigger_price,
        }

        self.modify_order(data)  # type: ignore

    def reverse_transaction_type(self, transaction_type):
        if transaction_type == "BUY":
            return "SELL"
        elif transaction_type == "SELL":
            return "BUY"
        raise Exception(f"unknown transaction type [{transaction_type}]")

    def place_large_orders(self, orders):
        order_ids = []
        orders = sorted(orders, key=lambda d: d["transaction_type"])
        for order in orders:
            exchange = order["exchange"]
            tradingsymbol = order["tradingsymbol"]
            quantity = order["quantity"]
            quantity_list = self.split_order(exchange, tradingsymbol, quantity)
            if len(quantity_list) > 10:
                raise InvalidArgumentException(
                    "Number of orders limit [10] exceeded, please reduce quantity"
                )

            for quantity in quantity_list:
                order_id = self.place_order(
                    tradingsymbol=tradingsymbol,
                    exchange=exchange,
                    quantity=quantity,
                    order_type="MARKET",
                    transaction_type=order["transaction_type"],
                    tag="move_stk",
                    product=order["product"],
                    price=0,
                    trigger_price=None,
                )
                order_ids.append(order_id)

        return order_ids

    def move_strike(
        self,
        tradingsymbol,
        exchange,
        product,
        transaction_type,
        quantity,
        strike_factor,
        new_strike,
    ):
        split_regex = r"([A-Z]+)(.{5})([0-9]+)(CE|PE)"

        underlying, expiry, strike, instrument_type = re.findall(
            split_regex, tradingsymbol
        )[0]
        strike = int(strike)

        config = self.underlying_config(underlying, expiry)
        strike_gap = config["strike_gap"]

        if strike_factor:
            new_strike = strike + (strike_gap * strike_factor)

        new_trading_symbol = f"{underlying}{expiry}{new_strike}{instrument_type}"

        orders = [
            {
                "tradingsymbol": tradingsymbol,
                "exchange": exchange,
                "transaction_type": self.reverse_transaction_type(transaction_type),
                "quantity": quantity,
                "product": product,
            },
            {
                "tradingsymbol": new_trading_symbol,
                "exchange": exchange,
                "transaction_type": transaction_type,
                "quantity": quantity,
                "product": product,
            },
        ]

        return self.place_large_orders(orders)

    def modify_orders_till_complete(
        self, orders_placed, sleep_time=10, max_modification_count=10
    ):
        modification_count = {}
        skip_first_sleep = True
        while 1:
            if not skip_first_sleep:
                time.sleep(sleep_time)
                skip_first_sleep = False

            orders = self.orders()

            orders = orders.filter(pl.col("order_id").is_in(orders_placed))
            orders = orders.filter(
                ~pl.col("status").is_in(["REJECTED", "CANCELLED", "COMPLETE"])
            )

            if len(orders) == 0:
                Constants.logger.info("ALL orders have been completed")
                break

            orders_dict = orders.to_dicts()
            for order in orders_dict:
                order_id = order["order_id"]
                if "PENDING" in order["status"]:
                    continue

                if order_id not in modification_count:
                    modification_count[order_id] = 0
                else:
                    modification_count[order_id] += 1

                ltp = self.get_ltp(order["exchange"], order["tradingsymbol"])

                market_protection = 0.02
                if modification_count[order_id] > 5:
                    market_protection = 0.05

                order["price"] = self.market_protection_price(
                    ltp, order["transaction_type"], market_protection=market_protection
                )

                order["order_type"] = "LIMIT"
                self.modify_order(order)  # type: ignore

                time.sleep(0.1)

                if modification_count[order_id] > max_modification_count:
                    order["order_type"] = "MARKET"
                    order["price"] = 0
                    Constants.logger.info("Placing MARKET order [{}]".format(order))
                    self.modify_order(order)  # type: ignore

                elif modification_count[order_id] > 20:
                    self.cancel_order(order_id)
                    Constants.logger.error(
                        "Max Modification Limit Exceeded : [{}]".format(order_id)
                    )

    # **
    # ** Generics
    # **

    def orders(self, tag: str | None = None, add_ltp: bool = True) -> pl.DataFrame:
        """Return user orders from the specified broker

        Args:
            tag (str | None, optional): orders tag. Defaults to None.
            add_ltp (bool, optional): want LTP or not. Defaults to True.

        Raises:
            NotImplementedError: if Function Not Implemented by Sub class

        Returns:
            pd.DataFrame: returns user orders
        """
        raise NotImplementedError

    def positions(self, drop_cnc: bool = True) -> pl.DataFrame:
        """Returns user position from the specified broker

        Args:
            drop_cnc (bool, optional): CNC positions to be dropped or not. Defaults to True.

        Raises:
            NotImplementedError: if Function Not Implemented by Sub class

        Returns:
            pd.DataFrame: returns user positions
        """
        raise NotImplementedError

    def modify_order(self, order_to_modify: ModifyOrderRequest) -> str:
        """Modifies Existing Order place on exchange

        Args:
            order_to_modify (OrderToModify): Order to be modified

        Raises:
            NotImplementedError: if Function Not Implemented by Sub class

        Returns:
            str: returns order_id
        """
        raise NotImplementedError

    def place_order(
        self,
        tradingsymbol: str,
        exchange: ExchangeType,
        quantity: int,
        order_type: OrderType,
        transaction_type: TransactionType,
        tag: str | None,
        product: ProductType,
        price: float,
        trigger_price: float | None = None,
    ) -> str:
        """Function for Place Order to exchange

        Args:
            tradingsymbol (str): TradingSymbol to place order
            exchange (ExchangeType): exchange
            quantity (int): number
            order_type (OrderType): order Type
            transaction_type (TransactionType): Transaction Type
            tag (str | None): order tag
            product (ProductType): product
            price (float): price
            trigger_price (float | None): trigger price

        Raises:
            NotImplementedError: if Function Not Implemented by Sub class

        Returns:
            str: Return orders_id
        """
        raise NotImplementedError

    def get_ltp(self, exchange: ExchangeType, tradingsymbol: str) -> float:
        """_summary_

        Args:
            exchange (ExchangeType): Exchange
            tradingsymbol (str): Tradingsymbol

        Raises:
            NotImplementedError: if Function Not Implemented by Sub class

        Returns:
            float: Returns ltp of the tradingsymbol
        """
        raise NotImplementedError

    def cancel_order(self, order_id: str) -> None:
        """Cancels order for the given order_id

        Args:
            order_id (str): Order ID

        Raises:
            NotImplementedError: if Function Not Implemented by Sub class
        """
        raise NotImplementedError

    def profile(self) -> UserBrokerProfileResponse:
        """Returns user broker profile

        Raises:
            NotImplementedError: if Function Not Implemented by Sub class

        Returns:
            UserBrokerProfile: contains exchange user_id full_name and email
        """
        raise NotImplementedError

    def holdings(self) -> pl.DataFrame:
        """Returns User Broker Holdings

        Raises:
            NotImplementedError: if Function Not Implemented by Sub class

        Returns:
            UserBrokerProfile: contains exchange user_id full_name and email
        """
        raise NotImplementedError
