from queue import Queue
import traceback
import json

import pandas as pd
import polars as pl
import pickle
import codecs
from datetime import datetime

from quantplay.broker.generics.broker import Broker
from quantplay.model.broker.generics import (
    ExchangeType,
    ModifyOrderRequest,
    TransactionType,
    UserBrokerProfileResponse,
)
from quantplay.utils.constant import Constants, timeit, OrderType
from quantplay.broker.xts_utils.Connect import XTSConnect
from quantplay.utils.pickle_utils import PickleUtils, InstrumentData
from quantplay.exception.exceptions import (
    InvalidArgumentException,
    RetryableException,
    retry_exception,
)
from quantplay.broker.xts_utils.InteractiveSocketClient import OrderSocket_io

from quantplay.exception.exceptions import TokenException
from retrying import retry
import requests
from urllib3.exceptions import InsecureRequestWarning
from quantplay.model.broker.generics import OrderType as NewOrderType
from quantplay.model.broker.generics import ProductType as NewProductType

# -# Suppress only the single warning from urllib3 needed.
requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)  # type: ignore


class XTS(Broker):
    source = "WebAPI"

    @timeit(MetricName="XTS:__init__")
    def __init__(
        self,
        root_url,
        api_key=None,
        api_secret=None,
        md_api_key=None,
        md_api_secret=None,
        order_updates: Queue | None = None,
        wrapper=None,
        md_wrapper=None,
        ClientID=None,
        load_instrument=True,
    ):
        super(XTS, self).__init__()
        self.order_updates: Queue | None = order_updates
        self.root_url = root_url

        try:
            if wrapper:
                self.set_wrapper(wrapper, md_wrapper)
                self.ClientID = ClientID
                self.wrapper.root = root_url
                self.md_wrapper.root = root_url
            else:
                self.login(api_key, api_secret, md_api_key, md_api_secret)
        except Exception as e:
            print(traceback.print_exc())
            raise e

        if load_instrument:
            self.load_instrument()

    def set_wrapper(self, serialized_wrapper, serialized_md_wrapper):
        self.wrapper = pickle.loads(codecs.decode(serialized_wrapper.encode(), "base64"))
        self.md_wrapper = pickle.loads(
            codecs.decode(serialized_md_wrapper.encode(), "base64")
        )

    def load_instrument(self, file_name: str | None = None) -> None:
        # TODO: Check for Futures
        try:
            self.symbol_data = InstrumentData.get_instance().load_data("xts_instruments")  # type: ignore
            Constants.logger.info("[LOADING_INSTRUMENTS] loading data from cache")
        except Exception:
            instruments = pd.read_csv(
                "https://quantplay-public-data.s3.ap-south-1.amazonaws.com/symbol_data/instruments.csv"
            )
            instruments = instruments.to_dict("records")
            self.symbol_data = {}

            for instrument in instruments:
                exchange = instrument["exchange"]
                tradingsymbol = instrument["tradingsymbol"]
                # NIFTY 08JUN2023 PE 17850 <- NIFTY2360817850PE
                # 2023-06-27 -> 08JUN2023
                # For FUTURES : EURINR23AUGFUT -> EURINR 23AUG2023 FUT

                ins_type = instrument["instrument_type"]
                name = instrument["name"]
                if ins_type in ["CE", "PE"]:
                    expiry = datetime.strftime(
                        datetime.strptime(str(instrument["expiry"]), "%Y-%m-%d"),
                        "%d%b%Y",
                    ).upper()
                    strike = str(instrument["strike"]).rstrip("0")
                    if strike[-1] == ".":
                        strike = strike[:-1]
                    instrument["broker_symbol"] = f"{name} {expiry} {ins_type} {strike}"
                elif ins_type == "FUT":
                    expiry = datetime.strftime(
                        datetime.strptime(str(instrument["expiry"]), "%Y-%m-%d"),
                        "%d%b%Y",
                    ).upper()
                    instrument["broker_symbol"] = f"{name} {expiry}"
                else:
                    instrument["broker_symbol"] = tradingsymbol

                self.symbol_data["{}:{}".format(exchange, tradingsymbol)] = instrument

            PickleUtils.save_data(self.symbol_data, "xts_instruments")
            Constants.logger.info("[LOADING_INSTRUMENTS] loading data from server")
        self.initialize_broker_symbol_map()

    def login(self, api_key, api_secret, md_api_key, md_api_secret):
        try:
            self.wrapper = XTSConnect(
                apiKey=api_key,
                secretKey=api_secret,
                root=self.root_url,
            )
            xt_core_response = self.wrapper.interactive_login()
            self.md_wrapper = XTSConnect(
                apiKey=md_api_key,
                secretKey=md_api_secret,
                root=self.root_url,
            )
            md_response = self.md_wrapper.marketdata_login()
            if "type" not in xt_core_response or xt_core_response["type"] != "success":
                print(f"api login response {xt_core_response}")
                raise TokenException("Api key credentials are incorrect")
            if "type" not in md_response or md_response["type"] != "success":
                print(f"market data login response {md_response}")
                raise TokenException("Market data api credentials are invalid")
            self.ClientID = xt_core_response["result"]["userID"]
        except TokenException:
            raise
        except Exception:
            raise InvalidArgumentException("Invalid api key/secret")

    def handle_exception(self, response):
        if "data" in response and "description" in response["data"]:
            data = response["data"]
            if "max limit" in data["description"].lower():
                user_id = self.profile()["user_id"]
                print("Rate limit problem")
                raise RetryableException(f"{user_id}: Request limit exceeded")
        if (
            "description" in response
            and "Authorization not found" in response["description"]
        ):
            raise TokenException(response["description"])
        if "type" in response and response["type"] == "error":
            raise Exception("[XTS_Error]: " + response["description"])

    @retry(
        wait_exponential_multiplier=3000,
        wait_exponential_max=10000,
        stop_max_attempt_number=3,
        retry_on_exception=retry_exception,
    )
    def margins(self) -> dict:
        # TODO: Edit For Dealers

        api_response = self.wrapper.get_balance(clientID=self.ClientID)
        self.handle_exception(api_response)

        if not api_response:
            return {
                "margin_used": 0,
                "margin_available": 0,
            }
        api_response = api_response["result"]["BalanceList"][0]["limitObject"]
        return {
            # TODO: Get PNL
            "pnl": 0,
            "margin_used": api_response["RMSSubLimits"]["marginUtilized"],
            "margin_available": api_response["RMSSubLimits"]["netMarginAvailable"],
        }

    def account_summary(self):
        # TODO: Edit For Dealers
        margins = self.margins()
        margins["pnl"] = float(self.positions()["pnl"].sum())
        return margins

    @retry(
        wait_exponential_multiplier=3000,
        wait_exponential_max=10000,
        stop_max_attempt_number=3,
        retry_on_exception=retry_exception,
    )
    def profile(self) -> UserBrokerProfileResponse:
        api_response = self.wrapper.get_profile(self.ClientID)
        self.handle_exception(api_response)
        api_response = api_response["result"]

        response: UserBrokerProfileResponse = {
            "user_id": api_response["ClientId"],
            "full_name": api_response["ClientName"],
            "segments": api_response["ClientExchangeDetailsList"],
        }

        return response

    @retry(
        wait_exponential_multiplier=3000,
        wait_exponential_max=10000,
        stop_max_attempt_number=3,
        retry_on_exception=retry_exception,
    )
    def orders(self, tag: str | None = None, add_ltp: bool = True) -> pl.DataFrame:
        api_response = self.wrapper.get_order_book(self.ClientID)
        self.handle_exception(api_response)

        api_response = api_response["result"]
        if api_response is None or len(api_response) == 0:
            return pl.DataFrame(schema=self.orders_schema)

        orders_df = pl.DataFrame(api_response)

        if len(orders_df) == 0:
            return orders_df

        orders_df = orders_df.rename(
            {
                "TradingSymbol": "tradingsymbol",
                "ClientID": "user_id",
                "AppOrderID": "order_id",
                "OrderStatus": "status",
                "ExchangeSegment": "exchange",
                "OrderPrice": "price",
                "OrderType": "order_type",
                "OrderSide": "transaction_type",
                "OrderAverageTradedPrice": "average_price",
                "OrderGeneratedDateTime": "order_timestamp",
                "OrderQuantity": "quantity",
                "CumulativeQuantity": "filled_quantity",
                "LeavesQuantity": "pending_quantity",
                "ProductType": "product",
                "OrderStopPrice": "trigger_price",
                "OrderUniqueIdentifier": "tag",
                "ExchangeInstrumentID": "token",
                "OrderCategoryType": "variety",
            }
        )
        orders_df = orders_df.with_columns(
            pl.struct(["tradingsymbol"])
            .map_elements(
                lambda x: self.get_quantplay_symbol(x["tradingsymbol"]),
                return_dtype=pl.String,
            )
            .alias("tradingsymbol")
        )
        if add_ltp:
            positions = self.positions()
            positions = positions.sort("product").group_by("tradingsymbol").head(1)

            orders_df = orders_df.join(
                positions.select(["tradingsymbol", "ltp"]), on="tradingsymbol", how="left"
            )

        orders_df = orders_df.with_columns(
            pl.col("filled_quantity").cast(pl.Int64).alias("filled_quantity"),
            pl.col("order_id").cast(pl.String).alias("order_id"),
        )
        orders_df = orders_df.with_columns(
            pl.when(pl.col("average_price") == "")
            .then(0)
            .otherwise(pl.col("average_price"))
            .cast(pl.Float64)
            .alias("average_price")
        )
        orders_df = orders_df.with_columns(
            (
                pl.col("ltp") * pl.col("filled_quantity")
                - pl.col("average_price") * pl.col("filled_quantity")
            ).alias("pnl")
        )

        orders_df = orders_df.with_columns(
            pl.when(pl.col("transaction_type") == "SELL")
            .then(-pl.col("pnl"))
            .otherwise(pl.col("pnl"))
            .alias("pnl")
        )

        orders_df = orders_df.with_columns(
            pl.col("order_timestamp")
            .str.strptime(pl.Datetime, "%d-%m-%Y %H:%M:%S")
            .alias("order_timestamp"),
            pl.col("LastUpdateDateTime")
            .str.strptime(pl.Datetime, "%d-%m-%Y %H:%M:%S")
            .alias("update_timestamp"),
        )

        # TODO add support for BSE
        orders_df = orders_df.with_columns(
            pl.when(pl.col("exchange") == "NSECM")
            .then(pl.lit("NSE"))
            .when(pl.col("exchange") == "NSEFO")
            .then(pl.lit("NFO"))
            .otherwise(pl.col("exchange"))
            .alias("exchange"),
            pl.when(pl.col("status") == "Rejected")
            .then(pl.lit("REJECTED"))
            .when(pl.col("status") == "Cancelled")
            .then(pl.lit("CANCELLED"))
            .when(pl.col("status") == "Filled")
            .then(pl.lit("COMPLETE"))
            .when(pl.col("status") == "New")
            .then(pl.lit("OPEN"))
            .otherwise(pl.col("status"))
            .alias("status"),
            pl.when(pl.col("order_type") == pl.lit("Limit"))
            .then(pl.lit("LIMIT"))
            .when(pl.col("order_type") == pl.lit("StopLimit"))
            .then(pl.lit("SL"))
            .when(pl.col("order_type") == pl.lit("Market"))
            .then(pl.lit("MARKET"))
            .otherwise(pl.col("order_type"))
            .alias("order_type"),
        )

        orders_df = orders_df.with_columns(
            pl.lit("").alias("status_message"), pl.lit("").alias("status_message_raw")
        )
        if tag:
            orders_df = orders_df.filter(pl.col("tag") == tag)

        return orders_df[list(self.orders_schema.keys())].cast(self.orders_schema)

    @retry(
        wait_exponential_multiplier=3000,
        wait_exponential_max=10000,
        stop_max_attempt_number=3,
    )
    @timeit(MetricName="XTS:holdings")
    def holdings(self):
        return pl.DataFrame(schema=self.holidings_schema)

    @retry(
        wait_exponential_multiplier=3000,
        wait_exponential_max=10000,
        stop_max_attempt_number=3,
        retry_on_exception=retry_exception,
    )
    def positions(self, drop_cnc: bool = True) -> pl.DataFrame:
        # TODO: get sell_value, buy_value
        api_response = self.wrapper.get_position_daywise(self.ClientID)
        self.handle_exception(api_response)

        api_response = api_response["result"]["positionList"]
        positions_df = pl.DataFrame(api_response)

        if len(positions_df) == 0:
            return pl.DataFrame(schema=self.positions_schema)

        positions_df = positions_df.rename(
            {
                "TradingSymbol": "tradingsymbol",
                "ExchangeSegment": "exchange",
                "OpenBuyQuantity": "buy_quantity",
                "OpenSellQuantity": "sell_quantity",
                "Quantity": "quantity",
                "SumOfTradedQuantityAndPriceBuy": "buy_value",
                "SumOfTradedQuantityAndPriceSell": "sell_value",
                "ProductType": "product",
                "ExchangeInstrumentId": "token",
            }
        )

        positions_df = positions_df.with_columns(
            pl.when(pl.col("exchange") == "NSECM")
            .then(pl.lit("NSE"))
            .when(pl.col("exchange") == "NSEFO")
            .then(pl.lit("NFO"))
            .otherwise(pl.col("exchange"))
            .alias("exchange")
        )

        positions_df = positions_df.with_columns(
            (pl.col("exchange") + ":" + pl.col("token")).alias("exchange_symbol")
        )

        symbols = (
            positions_df.select(pl.col("exchange_symbol").unique()).to_series().to_list()
        )
        symbol_ltps = self.get_ltps(symbols)

        positions_df = positions_df.with_columns(
            pl.col("token")
            .cast(pl.Int64)
            .replace_strict(symbol_ltps, default=0)
            .cast(pl.Float64)
            .alias("ltp")
        )

        positions_df = positions_df.with_columns(
            (
                pl.col("sell_value").cast(pl.Float64)
                - pl.col("buy_value").cast(pl.Float64)
            ).alias("pnl")
        )
        positions_df = positions_df.with_columns(
            (pl.col("pnl") + (pl.col("quantity").cast(pl.Float64) * pl.col("ltp"))).alias(
                "pnl"
            )
        )

        positions_df = positions_df.with_columns(
            pl.struct(["tradingsymbol"])
            .map_elements(
                lambda x: self.get_quantplay_symbol(x["tradingsymbol"]),
                return_dtype=pl.String,
            )
            .alias("tradingsymbol")
        )
        positions_df = positions_df.with_columns(
            pl.when(pl.col("tradingsymbol").str.slice(-2) == "PE")
            .then(pl.lit("PE"))
            .otherwise(pl.lit("CE"))
            .alias("option_type")
        )
        positions_df = positions_df.with_columns(
            pl.when(pl.col("exchange").is_in(["NFO", "BFO"]))
            .then(pl.col("option_type"))
            .otherwise(None)
            .alias("option_type")
        )
        positions_df = positions_df.with_columns(pl.lit(0).alias("average_price"))

        positions_df = positions_df[list(self.positions_schema.keys())].cast(
            self.positions_schema
        )

        positions_df = positions_df.with_columns(
            ((pl.col("buy_value") - pl.col("sell_value")) / pl.col("quantity")).alias(
                "average_price"
            )
        )

        return positions_df

    def get_quantplay_symbol(self, symbol):
        if symbol in self.broker_symbol_map:
            return self.broker_symbol_map[symbol]
        return symbol

    def get_ltps(self, symbols):
        instruments = [
            {
                "exchangeSegment": int(self.get_exchange_code(x.split(":")[0])),
                "exchangeInstrumentID": int(x.split(":")[1]),
            }
            for x in symbols
        ]
        api_response = self.md_wrapper.get_quote(
            Instruments=instruments,
            xtsMessageCode=1512,
            publishFormat="JSON",
        )

        if "type" in api_response and api_response["type"] == "error":
            raise TokenException(api_response["description"])
        api_response = api_response["result"]

        ltp_json = api_response["listQuotes"]

        ltp = [json.loads(x) for x in ltp_json]

        ltp = {x["ExchangeInstrumentID"]: x["LastTradedPrice"] for x in ltp}
        return ltp

    def get_exchange_code(self, exchange):
        exchange_code_map = {
            "NSE": 1,
            "NFO": 2,
            "BFO": 12,
            "BSE": 11,
            "NSECM": 1,
            "NSEFO": 2,
            "NSECD": 3,
            "BSECM": 11,
            "BSEFO": 12,
        }

        if exchange not in exchange_code_map:
            raise KeyError(
                f"INVALID_EXCHANGE: Exchange {exchange} not in ['NSE', 'NFO', 'NSECD', 'BSECM', 'BSEFO']"
            )

        return exchange_code_map[exchange]

    def get_exchange_name(self, exchange):
        exchange_code_map = {
            "NSE": "NSECM",
            "NFO": "NSEFO",
            "BFO": "BSEFO",
            "NSECD": "NSECD",
            "BSECM": "BSECM",
            "BSEFO": "BSEFO",
        }

        if exchange not in exchange_code_map:
            raise KeyError(
                f"INVALID_EXCHANGE: Exchange {exchange} not in ['NSE', 'NFO', 'NSECD', 'BSECM', 'BSEFO']"
            )

        return exchange_code_map[exchange]

    def get_ltp(self, exchange=None, tradingsymbol=None) -> float:
        exchange_code = self.get_exchange_code(exchange)
        exchange_token = self.symbol_data[f"{exchange}:{tradingsymbol}"]["exchange_token"]

        api_response = self.md_wrapper.get_quote(
            Instruments=[
                {
                    "exchangeSegment": exchange_code,
                    "exchangeInstrumentID": exchange_token,
                }
            ],
            xtsMessageCode=1512,
            publishFormat="JSON",
        )["result"]

        ltp_json = api_response["listQuotes"][0]

        ltp = json.loads(ltp_json)["LastTradedPrice"]

        return ltp

    def place_order(
        self,
        tradingsymbol: str,
        exchange: ExchangeType,
        quantity: int,
        order_type: NewOrderType,
        transaction_type: TransactionType,
        tag: str | None,
        product: NewProductType,
        price: float,
        trigger_price: float | None = None,
    ) -> str:
        exchange_name = self.get_exchange_name(exchange)
        xts_order_type = self.get_order_type(order_type)

        exchange_token = self.symbol_data[f"{exchange}:{tradingsymbol}"]["exchange_token"]
        if trigger_price is None:
            trigger_price = 0

        api_response = self.wrapper.place_order(
            exchangeSegment=exchange_name,
            exchangeInstrumentID=exchange_token,
            orderType=xts_order_type,
            disclosedQuantity=0,
            orderQuantity=quantity,
            limitPrice=price,
            timeInForce="DAY",
            stopPrice=trigger_price,
            orderSide=transaction_type,
            productType=product,
            orderUniqueIdentifier=tag,
            clientID=self.ClientID,
        )
        Constants.logger.info(f"[XTS_PLACE_ORDER_RESPONSE] {api_response}")

        if api_response["type"] == "error":
            Constants.logger.info(f"[XTS_PLACE_ORDER_ERROR] {api_response}")

            raise Exception("[XTS_ERROR]: " + api_response["description"])

        return api_response["result"]["AppOrderID"]

    def cancel_order(self, order_id):
        orders = self.orders()

        order_data = orders.filter(pl.col("order_id").eq(str(order_id)))
        if len(order_data) == 0:
            raise InvalidArgumentException(f"Order [{order_id}] not found")
        order_data = order_data.to_dicts()[0]

        tag = order_data["tag"]

        api_response = self.wrapper.cancel_order(
            appOrderID=int(order_id),
            clientID=order_data["user_id"],
            orderUniqueIdentifier=tag,
        )

        if api_response["type"] == "error":
            Constants.logger.info(f"[XTS_CANCEL_ORDER_ERROR] {api_response}")

            raise Exception("[XTS_ERROR]: " + api_response["description"])

        return api_response["result"]["AppOrderID"]

    def get_order_type(self, order_type: NewOrderType) -> str:
        if order_type == OrderType.market:
            return "Market"
        elif order_type == OrderType.sl:
            return "StopLimit"
        elif order_type == OrderType.slm:
            return "StopMarket"
        elif order_type == OrderType.limit:
            return "Limit"

        return order_type

    def modify_order(self, order_to_modify: ModifyOrderRequest) -> str:
        order_id = order_to_modify.get("order_id", None)
        price = order_to_modify.get("price", None)
        trigger_price = order_to_modify.get("trigger_price", None)
        order_type = order_to_modify.get("order_type", None)
        tag = order_to_modify.get("tag", None)
        product = order_to_modify.get("product", None)

        orders = self.orders()
        order_data = orders.filter(pl.col("order_id").eq(order_id))
        if len(order_data) == 0:
            raise InvalidArgumentException(f"Order [{order_id}] not found")
        order_data = order_data.to_dicts()[0]

        price = price or order_data["price"]
        trigger_price = trigger_price or order_data["trigger_price"]
        order_type = order_type or order_data["order_type"]
        tag = tag or order_data["tag"]
        product = product or order_data["product"]

        quantity = order_data["quantity"]
        time_in_force = "DAY"
        disclosed_quantity = 0

        api_response = self.wrapper.modify_order(
            appOrderID=order_id,
            modifiedTimeInForce=time_in_force,
            modifiedDisclosedQuantity=disclosed_quantity,
            modifiedLimitPrice=price,
            modifiedOrderQuantity=quantity,
            modifiedOrderType=self.get_order_type(order_type),
            modifiedProductType=product,
            modifiedStopPrice=trigger_price,
            orderUniqueIdentifier=tag,
            clientID=self.ClientID,
        )

        if api_response["type"] == "error":
            Constants.logger.info(f"[XTS_MODIFY_ORDER_ERROR] {api_response}")

            raise Exception("[XTS_ERROR]: " + api_response["description"])

        return api_response["result"]["AppOrderID"]

    def modify_price(self, order_id, price: float, trigger_price=None, order_type=None):
        self.modify_order(
            {
                "order_id": str(order_id),
                "price": price,
                "trigger_price": trigger_price,
                "order_type": order_type,
            }
        )

    def stream_order_updates(self):
        socket = OrderSocket_io(
            userID=self.ClientID,
            token=self.wrapper.token,
            root_url=self.root_url,
        )
        socket.setup_event_listners(on_order=self.order_event_handler)
        socket.connect()

    def order_event_handler(self, order):
        order = json.loads(order)
        new_ord = {}

        try:
            new_ord["placed_by"] = order["ClientID"]
            new_ord["tag"] = order["ClientID"]
            new_ord["order_id"] = order["AppOrderID"]
            new_ord["exchange_order_id"] = order["ExchangeOrderID"]
            new_ord["exchange"] = order["ExchangeSegment"]
            new_ord["tradingsymbol"] = order["TradingSymbol"]

            if new_ord["exchange"] == "NSEFO":
                new_ord["exchange"] = "NFO"
            elif new_ord["exchange"] == "NSECM":
                new_ord["exchange"] = "NSE"

            if new_ord["exchange"] in ["NFO", "MCX"]:
                new_ord["tradingsymbol"] = self.broker_symbol_map[
                    new_ord["tradingsymbol"]
                ]

            new_ord["order_type"] = order["OrderType"].upper()
            new_ord["product"] = order["ProductType"].upper()
            new_ord["transaction_type"] = order["OrderSide"].upper()
            new_ord["quantity"] = int(order["OrderQuantity"])

            if "OrderStopPrice" in order:
                new_ord["trigger_price"] = float(order["OrderStopPrice"])
            else:
                new_ord["trigger_price"] = None

            new_ord["price"] = float(order["OrderPrice"])
            new_ord["status"] = order["OrderStatus"].upper()

            if new_ord["status"] == "PENDINGNEW":
                new_ord["status"] = "TRIGGER PENDING"
            elif new_ord["status"] == "PENDINGCANCEL":
                new_ord["status"] = "PENDING"
            elif new_ord["status"] == "PENDINGREPLACE":
                new_ord["status"] = "TRIGGER PENDING"
            elif new_ord["status"] == "REPLACED":
                new_ord["status"] = "UPDATE"
            elif new_ord["status"] == "NEW":
                new_ord["status"] = "OPEN"
            elif new_ord["status"] == "FILLED":
                new_ord["status"] = "COMPLETE"

            if new_ord["order_type"].upper() == "MARKET":
                new_ord["order_type"] = OrderType.market
            elif new_ord["order_type"].upper() == "STOPLIMIT":
                new_ord["order_type"] = OrderType.sl
            elif new_ord["order_type"].upper() == "STOPMARKET":
                new_ord["order_type"] = OrderType.slm
            elif new_ord["order_type"].upper() == "LIMIT":
                new_ord["order_type"] = OrderType.limit

            if new_ord["order_type"] == "LIMIT" and new_ord["status"] == "UPDATE":
                new_ord["status"] = "OPEN"

            if new_ord["status"] == "TRIGGER PENDING" and new_ord["trigger_price"] == 0:
                return

            if self.order_updates:
                self.order_updates.put(new_ord)

        except Exception as e:
            print(e)
            Constants.logger.error("[ORDER_UPDATE_PROCESSING_FAILED] {}".format(e))
